/* faerion.c -- main start-up code, global variables;
 * Copyright (C) 2001-2002 ForestNet IRC Network
 * This is a free software; see file `license' for details.
 * $Id: faerion.c,v 1.22 2002/11/18 12:44:07 hex Exp $
 */

#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#ifdef _WIN32
# include <io.h>
# include <windows.h>
# include <process.h>
#else
# include <unistd.h>
# include <pwd.h>
#endif

#include "sys.h"
#include "struct.h"
#include "h.h"
#include "ircdlog.h"
#include "unicode.h"
#include "send.h"

#ifndef MIN
# define MIN(a,b) ((a) < (b) ? (a) : (b))
#endif

extern char *optarg;
extern int optind;

#ifdef NEED___progname
char *__progname = NULL;
#else
extern char *__progname;
#endif

int bootopt = 0, dorehash = 0, portnum = -1,
	debuglevel = -1, oper_closed = 0;

int lu_noninv, lu_inv, lu_serv, lu_oper, lu_unknown, lu_channel,
	lu_lu, lu_lserv, lu_clu, lu_mlu, lu_cglobalu, lu_mglobalu;

char *debugmode = "", *sbrk0, *configfile = "ircd.conf",
	*forced_server_name = NULL;

time_t NOW;
time_t nextconnect = 1;		/* time for next try_connections call */
time_t nextping = 1;		/* same as above for check_pings() */
time_t nextdnscheck = 0;	/* next time to poll dns to force timeouts */
time_t nextexpire = 1;		/* next expire run on the dns cache */
time_t lastlucheck = 0;

/* default options, overriden in ircd.conf */
char *network_name = NULL;
char *kline_address = NULL;
char *contact_url = NULL;
char *services_host = NULL;
char *server_mask = NULL;
char *ns_nick = NULL, *cs_nick = NULL, *ms_nick = NULL, *os_nick = NULL;
char *codepage_path = NULL;
char *opers_log = NULL;
char *clients_log = NULL;
int def_umode = 0;
int nickchange_delay = 0;
int show_connect_info = 0;
int force_motd = 0;

aClient me; /* That's me */
aClient *client = &me; /* Pointer to beginning of Client list */

char REPORT_DO_DNS[128], REPORT_FIN_DNS[128], REPORT_FIN_DNSC[128],
	REPORT_FAIL_DNS[128], REPORT_DO_ID[128], REPORT_FIN_ID[128],
	REPORT_FAIL_ID[128];

char buf[BUFSIZE];

static int doterminate = 0, termcode = 0;

extern void check_lusers(void);

#ifdef _WIN32
extern void nt_svc_maintain(char *, char *);
#endif

/*
 * Now find_kill is only called when a kline-related command is used:
 * [R]AKILL/[UN]KLINE/REHASH.  Very significant CPU usage decrease.
 * I made changes to evm_lusers every check_pings call to add new parameter.
 * -- Barubary
 */
time_t check_pings(time_t currenttime, int check_kills)
{
	aClient *cptr;
	int killflag, ping = 0, i, rflag = 0;
	time_t oldest = 0, timeout;

#ifdef TIMED_KLINES
	check_kills = 1;
#endif

	/* reversed scan to handle remaps */
	for (i = highest_fd; i >= 0; --i) {
		/* adjust if remapped */
		if (i > highest_fd)
			i = highest_fd;

		if (!(cptr = local[i]) || IsMe(cptr) || IsLog(cptr))
			continue;

		/* Note: No need to notify opers here. It's already done when
		  "FLAGS_DEADSOCKET" is set. */
		if (cptr->flags & FLAGS_DEADSOCKET) {
			exit_client(cptr, cptr, &me, "Dead socket");
			continue;
		}

		killflag = (check_kills) ? (IsPerson(cptr) ? find_kill(cptr) : 0) : 0;

		if (check_kills && !killflag && IsPerson(cptr))
			if (find_zap(cptr))
				killflag = 1;

#ifdef R_LINES_OFTEN
			rflag = IsPerson(cptr) ? find_restrict(cptr) : 0;
#endif

			ping = IsRegistered(cptr) ? get_client_ping(cptr)
				: CONNECTTIMEOUT;
			Debug((DEBUG_DEBUG, "c(%s)=%d p %d k %d r %d a %d",
				cptr->name, cptr->status, ping, killflag, rflag,
				currenttime - cptr->lasttime));

			/* Ok, so goto's are ugly and can be avoided here but this code
			   is already indented enough so I think its justified. -avalon */
			if (!killflag && !rflag && IsRegistered(cptr) &&
				(ping >= currenttime - cptr->lasttime))
					goto ping_timeout;

			/* If the server hasnt talked to us in 2*ping seconds
			   and it has a ping time, then close its connection.
			   If the client is a user and a KILL line was found
			   to be active, close this connection too. */
			if (killflag || rflag || ((currenttime - cptr->lasttime) >= (2 * ping) && (cptr->flags & FLAGS_PINGSENT)) ||
			   (!IsRegistered(cptr) && (currenttime - cptr->firsttime) >= ping)) {
				if (!IsRegistered(cptr) && (DoingDNS(cptr) || DoingAuth(cptr))) {
					if (cptr->authfd >= 0) {
#ifndef _WIN32
						close(cptr->authfd);
#else
						closesocket(cptr->authfd);
#endif
						cptr->authfd = -1;
						cptr->count = 0;
						*cptr->buffer = '\0';
					}
					if (show_connect_info) {
						if (DoingDNS(cptr))
							write(cptr->fd, REPORT_FAIL_DNS, strlen(REPORT_FAIL_DNS));
						else
							write(cptr->fd, REPORT_FAIL_ID, strlen(REPORT_FAIL_ID));
					}
					Debug((DEBUG_NOTICE, "DNS/AUTH timeout %s", get_client_name(cptr,TRUE)));
					del_queries((char *)cptr);
					ClearAuth(cptr);
					ClearDNS(cptr);
					SetAccess(cptr);
					cptr->firsttime = currenttime;
					cptr->lasttime = currenttime;
					continue;
				}

				if (IsServer(cptr) || IsConnecting(cptr) || IsHandshake(cptr)) {
					sendto_ops("No response from %s, closing link", get_client_name(cptr, FALSE));
					sendto_serv_butone(&me, ":%s GNOTICE :No response from %s, closing link", me.name, get_client_name(cptr, FALSE));
				}

				/* this is used for KILL lines with time restrictions on
				   them - send a messgae to the user being killed first. */
				if (killflag && IsPerson(cptr))
					sendto_ops("Kill line active for %s[%s@%s]", cptr->name, cptr->username, (*cptr->vhost) ? cptr->vhost : cptr->sockhost);

#if defined(R_LINES) && defined(R_LINES_OFTEN)
				if (IsPerson(cptr) && rflag)
					sendto_ops("Restricting %s, closing link.", get_client_name(cptr,FALSE));
#endif
				exit_client(cptr, cptr, &me, killflag ? gcomment : "Ping timeout");

				continue;
			}
			else if (IsRegistered(cptr) && (cptr->flags & FLAGS_PINGSENT) == 0) {
				/* if we havent PINGed the connection and we havent heard
				   from it in a while, PING it to make sure it is still
				   alive. */
				cptr->flags |= FLAGS_PINGSENT;

				/* not nice but does the job */
				cptr->lasttime = currenttime - ping;
				sendto_one(cptr, "PING :%s", me.name);
			}

ping_timeout:
		timeout = cptr->lasttime + ping;
		while (timeout <= currenttime)
			timeout += ping;
		if (timeout < oldest || !oldest)
			oldest = timeout;
	}
	if (!oldest || oldest < currenttime)
		oldest = currenttime + PINGFREQUENCY;

	return (oldest);
}


/* Scan through configuration and try new connections.  Returns the calendar
 * time when the next call to this function should be made latest. (No harm
 * done if this is called earlier or later...)
 * Borg88: Rewrited. No more stats c cycle.
 */
static time_t try_connections(time_t currenttime)
{
	aConfItem *aconf, *cconf, *con_conf;
	aClass *cltmp;
	aClient *cptr;
	int connecting, confrq, found, con_class = 0;
	time_t next = 0;

	connecting = FALSE;
	found = 0;

	for (aconf = conf; aconf; aconf = aconf->next) {
		/* Also when already connecting! (update holdtimes) --SRB */
		if (!(aconf->status & CONF_CONNECT_SERVER) || aconf->port <= 0)
			continue;
		if (aconf->flags & FLAG_TRIED) {
			found = 1;
			continue;
		}
		cltmp = Class(aconf);
		/*
		 * Skip this entry if the use of it is still on hold until
		 * future. Otherwise handle this entry (and set it on hold
		 * until next time). Will reset only hold times, if already
		 * made one successfull connection... [this algorithm is
		 * a bit fuzzy... -- msa >;) ]
		*/

		if ((aconf->hold > currenttime)) {
			if ((next > aconf->hold) || (next == 0))
				next = aconf->hold;
			continue;
		}

		confrq = get_con_freq(cltmp);
		aconf->hold = currenttime + confrq;

		/* Found a CONNECT config with port specified, scan clients
		 * and see if this server is already connected? */
		cptr = find_name(aconf->name, (aClient *)NULL);

		if (!connecting && !cptr && Links(cltmp) < MaxLinks(cltmp)) {
			/* Check connect rules to see if we're allowed to try */
			for (cconf = conf; cconf; cconf = cconf->next) {
				if ((cconf->status & CONF_CRULE)
				&& (match(cconf->host, aconf->name) == 0)) {
					if (crule_eval(cconf->passwd))
					break;
				}
			}

			if (!cconf) {
				con_class = Class(cltmp);
				con_conf = aconf;
				/* We connect only one at time... */
				connecting = TRUE;
			}
		}

		if (cptr || Links(cltmp)>=MaxLinks(cltmp))
			aconf->flags|=FLAG_TRIED;

		if((next > aconf->hold) || (next == 0))
			next = aconf->hold;
	}

	if (connecting) {
		if (connect_server(con_conf,
		(aClient *)NULL,(struct hostent *)NULL) == 0) {
			sendto_ops("Connection to %s[%s] activated.",
				con_conf->name, con_conf->host);
			con_conf->flags |= FLAG_TRIED;
			next = currenttime + 100;
		}
	}

	if (found && !next) {
		for (aconf = conf; aconf; aconf = aconf->next) {
			if (aconf->status & CONF_CONNECT_SERVER)
				aconf->flags &= ~FLAG_TRIED;
		}
		return 1;
	}

	return next;
}


void server_restart(char *message)
{
	putlog(LOG_INFO, "Child process exits for a restart, reason: %s.",
		message);

	sendto_ops("Server is restarting, reason: %s.", message);

	doterminate = 1;
	termcode = -1;
}


void server_terminate(char *message, ...)
{
	int i;
	aClient *acptr;
	char tmpbuf[1024];
	va_list vl;

	va_start(vl, message);
	vsnprintf(tmpbuf, sizeof(tmpbuf), message, vl);
	va_end(vl);
	tmpbuf[sizeof(tmpbuf) - 1] = '\0';

	putlog(LOG_INFO, "Child process exits (restart), reason: %s.", tmpbuf);

	for (i = 0; i <= highest_fd; i++) {
		if (!(acptr = local[i]))
			continue;

		if (IsServer(acptr))
			sendto_one(acptr, ":%s ERROR :Server terminated: %s.",
				me.name, tmpbuf);
		else if (IsClient(acptr))
			sendto_one(acptr, ":%s NOTICE %s :Server terminated: "
				"%s.", me.name, acptr->name, tmpbuf);
	}

	doterminate = 1;
	termcode = 0; /* normal shutdown */
}


void s_die(int raised)
{
	if (doterminate == 0)
		server_terminate("SIGTERM received");

	signal(raised, s_die);
}


static void s_dummy(int raised)
{
	signal(raised, s_dummy);
}


static void s_rehash(int raised)
{
	dorehash = 1;
	signal(raised, s_rehash);
}


static void s_restart(int raised)
{
	if (doterminate == 0)
		server_restart("received SIGINT");
}


static void setup_signals(void)
{
#ifdef SIGPIPE
	signal(SIGPIPE, s_dummy);
#endif
#ifdef SIGALRM
	signal(SIGALRM, s_dummy);
#endif
#ifdef SIGHUP
	signal(SIGHUP, s_rehash);
#endif
	signal(SIGTERM, s_die);
	signal(SIGINT, s_restart);
#ifdef SIGBREAK /* only under win32 so far */
	signal(SIGBREAK, s_rehash);
#endif
}


static void open_debugfile(int need)
{
#ifdef DEBUGMODE
	extern FILE *debugfile;

	debugfile = need ? fopen("debug.log", "w") : NULL;
#endif
}


static void print_usage(void)
{
	char pad[128];

	memset(pad, 0x20, sizeof(pad));
	pad[strlen(__progname) < sizeof(pad) ? strlen(__progname)
		: sizeof(pad) - 1] = '\0';

	printf("Usage: %s [-f ircd.conf] [-h server_name] [-p port_number]\n",
		__progname);

#ifndef _WIN32
	printf("       %s [-u user] [-g group] [-r chroot_path]\n", pad);
#else
	printf("       %s [-n service_name] [-k install|uninstall|configure]\n",
		pad);
#endif
	printf("       %s [-v verbosity] [-x debug_verbosity] [-Veq]\n", pad);

	printf(
	"Options:\n"
	"  -c           : stay attached to console (forces -e)\n"
	"  -e           : log to stderr instead of syslog\n"
	"  -f file      : specify an alternative configuration file\n"
#ifndef _WIN32
	"  -g group|id  : run as the specified group\n"
#endif
	"  -h name      : force a server name\n"
#ifdef _WIN32
	"  -k configure : change start-up configuration for the service\n"
	"  -k install   : install service\n"
	"  -k uninstall : uninstall service\n"
	"  -n name      : name NT service (defaults to \"faerion\")\n"
#endif
	"  -p number    : force port number to listen for server connections\n"
	"  -q           : quiet mode, equals to -v 0 (emergency errors)\n"
#ifndef _WIN32
	"  -r path      : chroot into the specified directory\n"
	"  -u user|id   : run as the specified user\n"
#endif
	"  -v level     : set logging verbosity (defaults to 4, warnings)\n"
	"  -V           : print version number and exit\n"
	"  -x level     : set debug.log verbosity (forces -v 9)\n"
	);
}


static void print_version(void)
{
	printf("Faerion IRC Daemon, version %s%s\n"
		"Mail bug reports and suggestions to <ircd@forestnet.org>.\n",
		get_ircd_version(),
#ifdef DEBUGMODE
		" (debug build)"
#else
		""
#endif
		);
}


/* main child process that actually does all the work.  parameters are passed
 * by the caller process using environment, which must take care about unused
 * ones so that nothing could be forged.  return values:
 *
 * 0 -- shutting down;
 * 1 -- could not open configuration file, fatal;
 *
 * all other values cause the process to be restarted;
 */
int child_main(void)
{
	char *val;
	time_t delay = 0, now;
	aConfItem *aconf;

#ifdef _WIN32
	WSADATA wsaData;
	if (WSAStartup(MAKEWORD(1, 1), &wsaData) != 0) {
		/* add something to EventLog */
		/* fprintf(stderr, "Could not initialize winsock.\n"); */
		return 0;
	}
#endif

	memset(&me, 0, sizeof(me));
#ifdef USE_CASETABLES
	setup_match();
#endif

	if ((val = getenv("faerion_parent_pid"))) {
		if ((val = getenv("faerion_bootopt")))
			bootopt = atoi(val);
		else
			bootopt = 0;

		if ((val = getenv("faerion_portnum")))
			portnum = atoi(val);
		else
			portnum = -1;

		if ((val = getenv("faerion_server_name")))
			strlcpy(me.name, val, sizeof(me.name));

		if ((val = getenv("faerion_debug_level")))
			debuglevel = atoi(val);
	}

	if (forced_server_name)
		strlcpy(me.name, forced_server_name, sizeof(me.name));

	open_debugfile(bootopt & BOOT_DEBUG);

	clear_client_hash_table();
	clear_channel_hash_table();
	clear_notify_hash_table();

	inittoken();
	initlists();
	initclass();
	initwhowas();
	initstats();

	me.port = portnum;
	init_sys();
	me.flags = FLAGS_LISTEN;
	me.fd = -1;

	if ((initconf(bootopt) == -1) || ((aconf = find_me()) == NULL))
		return 1;

	charset_refresh();

	if (portnum < 0) {
		if (aconf->port > 1023)
			portnum = aconf->port;
		else
			portnum = PORTNUM;
	}

	if (inetport(&me, aconf->passwd, portnum))
		return -1;

	setup_ping();
	get_my_name(&me, me.sockhost, sizeof(me.sockhost) - 1);

	if (me.name[0] == '\0')
		strlcpy(me.name, me.sockhost, sizeof(me.name));

	me.hopcount = 0;
	me.authfd = -1;
	me.confs = NULL;
	me.next = NULL;
	me.user = NULL;
	me.from = &me;
	SetMe(&me);
	make_server(&me);
	strcpy(me.serv->up, me.name);

	me.lasttime = me.since = me.firsttime = time(NULL);
	add_to_client_hash_table(me.name, &me);

	sprintf(REPORT_DO_DNS, ":%s %s", me.name, BREPORT_DO_DNS);
	sprintf(REPORT_FIN_DNS, ":%s %s", me.name, BREPORT_FIN_DNS);
	sprintf(REPORT_FIN_DNSC, ":%s %s", me.name, BREPORT_FIN_DNSC);
	sprintf(REPORT_FAIL_DNS, ":%s %s", me.name, BREPORT_FAIL_DNS);
	sprintf(REPORT_DO_ID, ":%s %s", me.name, BREPORT_DO_ID);
	sprintf(REPORT_FIN_ID, ":%s %s", me.name, BREPORT_FIN_ID);
	sprintf(REPORT_FAIL_ID, ":%s %s", me.name, BREPORT_FAIL_ID);

	check_class();
	setup_signals();

	while (1) {
		now = time(NULL);
		/* Run through the hashes and check lusers every second */
		if (lastlucheck < now) {
			check_lusers();
			lastlucheck = now;
		}

		/*
		 * We only want to connect if a connection is due,
		 * not every time through.  Note, if there are no
		 * active C lines, this call to Tryconnections is
		 * made once only; it will return 0. - avalon
		 */
		if (nextconnect && now >= nextconnect)
			nextconnect = try_connections(now);

		/*
		 * DNS checks. One to timeout queries, one for cache expiries.
		 */
		if (now >= nextdnscheck)
			nextdnscheck = timeout_query_list(now);
		if (now >= nextexpire)
			nextexpire = expire_cache(now);

		/*
		 * take the smaller of the two 'timed' event times as
		 * the time of next event (stops us being late :) - avalon
		 * WARNING - nextconnect can return 0!
		 */
		if (nextconnect)
			delay = MIN(nextping, nextconnect);
		else
			delay = nextping;
		delay = MIN(nextdnscheck, delay);
		delay = MIN(nextexpire, delay);
		delay -= now;

		/*
		 * Adjust delay to something reasonable [ad hoc values]
		 * (one might think something more clever here... --msa)
		 * We don't really need to check that often and as long
		 * as we don't delay too long, everything should be ok.
		 * waiting too long can cause things to timeout...
		 * i.e. PINGS -> a disconnection :( - avalon
		 */
		if (delay < 1)
			delay = 1;
		else
			delay = MIN(delay, TIMESEC);

		if (read_message(delay) != 0)
			Debug((DEBUG_DEBUG, "Got message(s)"));

		now = time(NULL);
		/*
		 * ...perhaps should not do these loops every time,
		 * but only if there is some chance of something
		 * happening (but, note that conf->hold times may
		 * be changed elsewhere--so precomputed next event
		 * time might be too far away... (similarly with
		 * ping times) --msa
		 */
		if (now >= nextping)
			nextping = check_pings(now, 0);

		if (dorehash) {
			rehash(&me, &me, 1);
			dorehash = 0;
		}

		if (doterminate != 0) {
			flush_connections(me.fd);
			sleep(1);
			return termcode;
		}

		/*
		 * Flush output buffers on all connections now if they
		 * have data in them (or at least try to flush) -avalon
		 */
		flush_connections(me.fd);
	}

}


void child_monitor(void)
{
	char tmpbuf[1024];

	sprintf(tmpbuf, "faerion_parent_pid=%d", getpid());
	if (putenv(tmpbuf))
		goto enverr;

	sprintf(tmpbuf, "faerion_bootopt=%d", bootopt);
	if (putenv(tmpbuf))
		goto enverr;

	sprintf(tmpbuf, "faerion_portnum=%d", portnum);
	if (putenv(tmpbuf))
		goto enverr;

	if (forced_server_name) {
		sprintf(tmpbuf, "faerion_server_name=%s", forced_server_name);
		if (putenv(tmpbuf))
			goto enverr;
	}

	if (bootopt & BOOT_DEBUG) {
		sprintf(tmpbuf, "faerion_debug_level=%d", debuglevel);
		if (putenv(tmpbuf))
			goto enverr;
	}

	write_pidfile();

	switch (child_main()) {
		case 1:
			putlog(LOG_EMERG, "error processing configuration "
				"file: %s", configfile);
			break;
		default:
			putlog(LOG_INFO, "child process quits.");
	}

	kill_pidfile();
	return;

enverr:
	fprintf(stderr, "Could not set environment variable: %s\n", tmpbuf);
	exit(1);
}


int main(int argc, char **argv)
{
	int ch;

#ifndef _WIN32
	char *want_root = NULL;
	uid_t want_uid = 0;
	gid_t want_gid = 0;
#else /* _WIN32 */
	char *svc_name = NULL, *svc_command = NULL;
#endif

#ifdef NEED___progname
	__progname = argv[0];
	if (strrchr(__progname, '/') != NULL)
		__progname = strrchr(__progname, '/') + 1;
	else if (strrchr(__progname, '\\') != NULL)
		__progname = strrchr(__progname, '\\') + 1;
#endif

#ifdef _WIN32
# define OPTIONS "n:k:"
#else
# define OPTIONS "u:g:r:"
#endif

	while ((ch = getopt(argc, argv, "qcVev:f:h:p:x:?" OPTIONS)) != -1) {
		switch (ch) {
			case 'V':
				print_version();
				exit(0);
			case 'v':
				log_verbosity = atoi(optarg);
				break;
			case 'q':
				log_verbosity = LOG_EMERG;
				break;
			case 'f':
				configfile = optarg;
				break;
			case 'c':
				bootopt |= BOOT_CONSOLE|BOOT_STDERR;
				break;
			case 'e':
				bootopt |= BOOT_STDERR;
				break;
#ifdef _WIN32
			case 'n':
				svc_name = optarg;
				break;
			case 'k':
				svc_command = optarg;
				break;
#else
			case 'u':
				if (!(want_uid = (uid_t)atoi(optarg))) {
					struct passwd *pw;
					if (pw = getpwnam(optarg)) {
						want_uid = pw->pw_uid;
						want_gid = pw->pw_gid;
					}
				}
				break;
			case 'g':
				want_gid = (gid_t)atoi(optarg);
				break;
			case 'r':
				want_root = optarg;
				break;
#endif
			case 'h':
				if (strchr(optarg, ' ')) {
					fprintf(stderr, "Server name must not "
						"contain spaces.\n");
					exit(1);
				}
				forced_server_name = optarg;
				break;
			case 'p': {
				int portarg;

				if ((portarg = atoi(optarg)) >= 1024)
					portnum = portarg;
				else {
					fprintf(stderr, "Port number must be "
						"greater than 1023.\n");
					exit(1);
				}
				break;
			}
			case 'x':
				log_verbosity = LOG_DEBUG;
				debuglevel = atoi(optarg);
				debugmode = debuglevel ? optarg : "0";
				bootopt |= BOOT_DEBUG;
				break;
			default:
				print_usage();
				exit(1);
		}
	}

	argc -= optind;
	argv += optind;

	umask(077);

#ifndef NEED_syslog
	openlog(__progname, LOG_PID|LOG_NDELAY, LOG_USER);
#endif

	if (bootopt & BOOT_CONSOLE)
		printf("Staying at console..\n");
	else if (daemon(1, 0)) {
		putlog(LOG_ERR, "Failed to detach from console, won't run.");
		return 1;
	}

#ifndef _WIN32
	if ((getuid() == 0) && (want_root != NULL)) {
		if (chroot(want_root)) {
			putlog(LOG_ERR, "failed to chroot: %s",
				strerror(errno));
			exit(1);
		}
		chdir("/");
	}

	if (want_gid && (setegid(want_gid) || setgid(want_gid))) {
		putlog(LOG_ERR, "failed to change group: %s",
			strerror(errno));
		exit(1);
	}

	if (want_uid && (seteuid(want_uid) || setuid(want_uid))) {
		putlog(LOG_ERR, "failed to change user: %s",
			strerror(errno));
		exit(1);
	}

	if (getuid() == 0 || getgid() == 0) {
		putlog(LOG_ERR, "I won't run as privileged user/group.");
		exit(1);
	}
#endif

#ifdef _WIN32
	/* this calls child_monitor() and exits the current thread, if
	 * the service was successfully started.
	 */
	if (svc_command)
		nt_svc_maintain(svc_name, svc_command);
#endif

	child_monitor();
	return 0;
}
