/* ircdpw.c -- generate MD5 passwords for use with the daemon,
 *             the passwords are FreeSec-compatible;
 * Copyright (C) 2001-2002 ForestNet IRC Network;
 * This is a free software; see file `license' for details.
 * $Id: ircdpw.c,v 1.3 2002/11/20 20:31:15 hex Exp $
 */

#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#ifdef _WIN32
# include <process.h>
# include <windows.h>
#else
# include <string.h> /* strchr() */
# include <strings.h>
# include <termios.h>
# include <unistd.h>
#endif

#include "ircdmd5.h"
#include "sys.h"

#ifndef NEED_arc4random
# define md5_srand() arc4random_stir()
# define md5_rand() arc4random()
#else
# define md5_srand() srand((unsigned int)time(0) * (unsigned int)getpid())
# define md5_rand() rand()
#endif

#ifndef _WIN32
struct termios oterm;
#endif

#ifdef NEED___progname
# define __progname (strchr(argv[0], '\\') ? strrchr(argv[0], '\\') : argv[0])
#else
extern char *__progname;
#endif


static unsigned char itoa64[] =
	"./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";


static void to64(char *s, unsigned int v, int n)
{
	while (--n >= 0) {
		*s++ = itoa64[v&0x3f];
		v >>= 6;
	}
}


static void echo_off(void)
{
#if defined(_WIN32)
	HANDLE hCon;
	DWORD mode;

	if ((hCon = GetStdHandle(STD_INPUT_HANDLE)) == INVALID_HANDLE_VALUE) {
		printf("Could not get console handle.\n");
		exit(1);
	}

	if (!GetConsoleMode(hCon, &mode)) {
		printf("Could not retreive console mode.\n");
		exit(1);
	}

	if (!SetConsoleMode(hCon, mode & ~ENABLE_ECHO_INPUT)) {
		printf("Could not disable echo.\n");
		exit(1);
	}
#else /* defined(_WIN32) */
	struct termios term;

	if (tcgetattr(STDIN_FILENO, &oterm) != 0) {
		printf("Could not get terminal status to disable echo.\n");
		exit(1);
	}

	memcpy(&term, &oterm, sizeof(term));
	term.c_lflag &= ~ECHO;
	tcsetattr(STDIN_FILENO, TCSAFLUSH, &term);
#endif
}


static void echo_on(void)
{
#ifdef _WIN32
	HANDLE hCon;
	DWORD mode;

	if (hCon = GetStdHandle(STD_INPUT_HANDLE)) {
		if (GetConsoleMode(hCon, &mode)) {
			mode |= ENABLE_ECHO_INPUT;
			SetConsoleMode(hCon, mode);
		}
	}
#else
	tcsetattr(STDIN_FILENO, TCSAFLUSH, &oterm);
#endif
}


static void sig_handler(int signal)
{
	printf("aborted by user.\n");
	echo_on();
	exit(1);
}


int main(int argc, char **argv)
{
	unsigned char salt[120], passwd[120], *p;

	if (argc > 1 && !strcmp(argv[1], "-?")) {
		printf("Faerion IRC Daemon, password generator.\n"
			"Reads password from standard input (no echo), "
			"returns a FreeSec-compatible MD5 hash.\n\n");
		printf("Usage: %s [salt]\n", __progname);
		exit(1);
	}

	signal(SIGINT, sig_handler);
#ifdef SIGBREAK
	signal(SIGBREAK, sig_handler);
#endif

	echo_off();
	printf("Enter string: ");
#ifdef _WIN32
	gets(passwd);
#else
	fgets(passwd, sizeof(passwd), stdin);
#endif
	echo_on();

	if (p = strchr(passwd, '\r'))
		*p = '\0';
	if (p = strchr(passwd, '\n'))
		*p = '\0';

	if (strlen(passwd) == 0) {
		printf("\nNo password was entered.\n");
		exit(1);
	}

	/* silly salt generator */
	if (argc > 1) {
		strncpy(salt, argv[1], sizeof(salt));
		salt[sizeof(salt) - 1] = '\0';
	} else {
		md5_srand();
		strcpy(salt, "$1$");
		to64(&salt[3], md5_rand(), 4);
		to64(&salt[7], md5_rand(), 4);
		strcpy(salt + 11, "$");
	}

	printf("\n%s\n", md5crypt(passwd, salt));

	return 0;
}
