/* m_close.c -- closing non-oper connections
 * Copyright (C) 2001-2002, ForestNet IRC Network
 * This is a free software; see file `license' for details.
 * $Id: m_close.c,v 1.12 2002/11/18 12:44:07 hex Exp $
 */

#include <stdarg.h>

#include "struct.h"
#include "send.h"
#include "numeric.h"
#include "h.h"
#include "msg.h"

/* by default this only closes unknown connections.  options are available:
 * CLOSE users -- close all connections except the opers
 * CLOSE all -- close all connections except yourself
 */
int m_close(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	aClient *acptr;
	int i, closed, mode;

	if (check_registered_user(sptr) || !MyClient(sptr))
		return 0;

	if (!OPIsAdmin(sptr)) {
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
		return 0;
	}

	mode = 0;

	if (parc > 1) {
		if (!strcasecmp(parv[1], "users"))
			mode = 1;
		else if (!strcasecmp(parv[1], "all"))
			mode = 2;
		else {
			send_syntax(sptr, MSG_CLOSE, "CLOSE [all|users]");
			return 0;
		}
	}

	for (closed = 0, i = highest_fd; i; --i) {
		if (!(acptr = local[i]) || !IsPerson(acptr) || acptr == sptr)
			continue;

		if (IsAnOper(acptr) && mode < 2)
			continue;

		if (IsClient(acptr) && mode < 1)
			continue;

		if (!IsUnknown(acptr) && !IsConnecting(acptr)
			&& !IsHandshake(acptr) && mode == 0)
			continue;

		sendto_one(sptr, rpl_str(RPL_CLOSING), me.name, parv[0],
			get_client_name(acptr, TRUE), acptr->status, i);
		exit_client(acptr, acptr, acptr, "Oper Closing");
		++closed;
	}

	if (mode)
		oper_closed = 1;

	sendto_one(sptr, rpl_str(RPL_CLOSEEND), me.name, parv[0], closed);
	return 0;
}
