/* m_gline.c -- broadcast a K-Line to all servers
 * Copyright (C) 2001-2002, ForestNet IRC Network
 * This is a free software; see file `license' for details.
 * $Id: m_gline.c,v 1.11 2002/11/18 12:44:08 hex Exp $
 */

#include <string.h>
#include <time.h>

#include "struct.h"
#include "numeric.h"
#include "msg.h"
#include "h.h"
#include "send.h"

int m_gline(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	aClient *acptr;
	char *hosttemp, host[128], name[128];

	if (check_registered(sptr))
		return 0;

	if (MyClient(sptr) && !OPIsOper(sptr)) {
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
		return 0;
	}

	if (parc < 2) {
		send_syntax(sptr, MSG_GLINE,
			"GLINE nick|[user@]host [:reason ...]");
		return 0;
	}

	if (MyClient(sptr) && (acptr = find_person(parv[1], NULL))) {
		/* strcpy(name, acptr->user->username); */
		strcpy(name, "*");
		strcpy(host, acptr->user->host);
	}
	else {
		if (strlen(parv[1]) > 127) {
			sendto_one(sptr, "NOTICE %s :Bad parameters", parv[0]);
			return 0;
		}

		memset(name, 0, sizeof(name));
		if (hosttemp = strchr(parv[1], '@')) {
			++hosttemp;
			strncpy(name, parv[1], hosttemp - 1 - parv[1]);
		}
		else {
			hosttemp = parv[1];
			strcpy(name, "*");
		}

		strcpy(host, hosttemp);
		if (name[0] == '\0' || host[0] == '\0') {
			sendto_one(sptr, "NOTICE %s :If you're going to add a "
				"userhost, at LEAST specify both fields",
				parv[0]);
			return 0;
		}

		if (!strcmp(host, "*")) {
			sendto_one(sptr, "NOTICE %s :What a sweeping K-Line.  "
				"If only your admin knew you tried that..",
				parv[0]);
			return 0;
		}
	}

	if (parc > 2)
		sendto_serv_butone((IsServer(cptr)) ? cptr : 0,
			":%s GLINE %s@%s :%s", parv[0], name, host, parv[2]);
	else
		sendto_serv_butone((IsServer(cptr)) ? cptr : 0,
			":%s GLINE %s@%s", parv[0], name, host);

	sendto_ops("%s added a temp K-Line %s@%s %s", parv[0], name, host,
		parv[2] ? parv[2] : "");

	add_temp_conf(CONF_KILL, host, parv[2], name, 0, 0, 1);
	check_pings(time(NULL), 1);

	return 0;
}
