/* m_user.c -- user registration message;
 * Copyright (C) 2001-2002, ForestNet IRC Network
 * This is a free software; see file `license' for details.
 * $Id: m_user.c,v 1.4 2002/11/18 12:44:08 hex Exp $
 */

#include <string.h>

#include "sys.h"
#include "struct.h"
#include "send.h"
#include "numeric.h"
#include "h.h"

#define UFLAGS (UMODE_INVISIBLE|UMODE_WALLOP|UMODE_SERVNOTICE)

extern int register_user(aClient *, aClient *, char *, char *);

/*
 * m_user
 *     parv[0] = sender prefix
 *     parv[1] = username (login name, account)
 *     parv[2] = client host name (used only from other servers)
 *     parv[3] = server host name (used only from other servers)
 *     parv[4] = users real name info
 */
int m_user(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	unsigned int sstamp = 0;
	char *username, *host, *server, *realname;
	anUser *user;

	if (IsServer(cptr) && !IsUnknown(sptr))
		return 0;

	if (parc > 2 && (username = strchr(parv[1], '@')))
		*username = '\0';

	if (parc < 5 || *parv[1] == '\0' || *parv[2] == '\0'
	|| *parv[3] == '\0' || *parv[4] == '\0') {
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name,
			parv[0], "USER");
		if (IsServer(cptr))
			sendto_ops("bad USER param count for %s from %s",
				parv[0], get_client_name(cptr, FALSE));
		else
			return 0;
	}

	/* Copy parameters into better documenting variables  */

	username = (parc < 2 || BadPtr(parv[1])) ? "<nouser>" : parv[1];
	host = (parc < 3 || BadPtr(parv[2])) ? "<nohost>" : parv[2];
	server = (parc < 4 || BadPtr(parv[3])) ? "<noserver>" : parv[3];

	/* This we can remove as soon as all servers have upgraded. */

	if (parc == 6 && IsServer(cptr)) {
		if (isdigit(*parv[4]))
			sstamp = atol(parv[4]);
		realname = (BadPtr(parv[5])) ? "<bad-realname>" : parv[5];
	}
	else
		realname = (BadPtr(parv[4])) ? "<bad-realname>" : parv[4];

	user = make_user(sptr);

	if (!MyConnect(sptr)) {
		if (sptr->srvptr == NULL)
			sendto_ops("WARNING, user %s introduced as being on "
				"non-existant server %s.", sptr->name, server);
		strlcpy(user->host, host, sizeof(user->host));
		goto user_finish;
	}

	if (!IsUnknown(sptr)) {
		sendto_one(sptr, err_str(ERR_ALREADYREGISTRED),
			me.name, parv[0]);
		return 0;
	}

	sptr->umodes |= (UFLAGS & atoi(host));
	sptr->umodes |= def_umode;

	strlcpy(user->host, host, sizeof(user->host));

user_finish:
	user->servicestamp = sstamp;
	strlcpy(sptr->info, realname, sizeof(sptr->info));

	if (sptr->name[0])
		return register_user(cptr, sptr, sptr->name, username);
	else
		strlcpy(sptr->user->username, username, USERLEN + 1);

	return 0;
}
