/* packet.c -- reading packets;
 * Copyright (C) 2001-2002, ForestNet IRC Network
 * This is a free software; see file `license' for details.
 * $Id: packet.c,v 1.10 2002/11/18 12:44:08 hex Exp $
 */

#include <time.h>

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "msg.h"
#include "h.h"
#include "zlink.h"
#include "send.h"

extern time_t NOW;

/* inittoken
 * Cheat here, blah. Build the lookup tables from msgtab's,
 * call them msgmap's.  Called in main() with other inits.
 * Yes, I know this is not the right module, but I said I cheat ;) */
void inittoken(void)
{
	int loopy;
	int final;

	/* Find the zero-entry */
	for (final = 0; msgtab[final].cmd; final++);

	/* Point all entries to it */
	for (loopy = 0; loopy < 256; loopy++)
		msgmap[loopy] = &msgtab[final];

	/* Build references to existing commands */
	for (loopy = 0; msgtab[loopy].cmd; loopy++)
		msgmap[msgtab[loopy].token[0]] = &msgtab[loopy];
}


/* dopacket
 *	cptr - pointer to client structure for which the buffer data
 *		applies.
 *	buffer - pointer to the buffer containing the newly read data
 *	length - number of valid bytes of data in the buffer
 *
 * Note:
 *	It is implicitly assumed that dopacket is called only
 *	with cptr of "local" variation, which contains all the
 *	necessary fields (buffer etc..)
 */

int dopacket(aClient *cptr, char *buffer, int length)
{
	int nlen;
	char *ch1, *ch2, *cptrbuf = cptr->buffer, *nbuf = NULL;
	aClient *acpt = cptr->acpt;

	me.receiveB += length; /* Update bytes received */
	cptr->receiveB += length;
	if (cptr->receiveB > 1023) {
		cptr->receiveK += (cptr->receiveB >> 10);
		cptr->receiveB &= 0x03ff; /* 2^10 = 1024, 3ff = 1023 */
	}

	if (acpt != &me) {
		acpt->receiveB += length;
		if (acpt->receiveB > 1023) {
			acpt->receiveK += (acpt->receiveB >> 10);
			acpt->receiveB &= 0x03ff;
		}
	}

	else if (me.receiveB > 1023) {
		me.receiveK += (me.receiveB >> 10);
		me.receiveB &= 0x03ff;
	}

	Debug((DEBUG_INFO, "packet from %s (%d bytes)", cptr->name, length));

zcontinue:
	ch1 = cptrbuf + cptr->count;
	ch2 = buffer;

	if (ZipIn(cptr)) {
		int err;
		ch2 = zip_input(cptr->serv->zip_in, ch2, &length, &err, &nbuf, &nlen);

		if (length == -1) {
			sendto_realops("Zipin error[1] for %s: (%d) %s\n", cptr->name, err, ch2);
			return exit_client(cptr, cptr, &me, "fatal error in compressed input");
		}
	}

	while (--length >= 0) {
		char g=(*ch1 = *ch2++);
		/* Yuck.  Stuck.  To make sure we stay backward compatible,
		 * we must assume that either CR or LF terminates the message
		 * and not CR-LF.  By allowing CR or LF (alone) into the body
		 * of messages, backward compatibility is lost and major
		 * problems will arise. - Avalon */

		if (g == '\r' || g == '\n') {
			if (ch1 == cptr->buffer)
				continue; /* Skip extra LF/CR's */
			*ch1 = '\0';

			me.receiveM += 1; /* Update messages received */
			cptr->receiveM += 1;
			if (cptr->acpt != &me)
				cptr->acpt->receiveM += 1;

			/* ...just in case parse returns with FLUSH_BUFFER
			 * without removing the structure pointed by
			 * cptr... --msa */
			cptr->count = 0;

			switch (parse(cptr, cptr->buffer, ch1, msgtab)) {
				case FLUSH_BUFFER:
					/* FLUSH_BUFFER means actually that
					 * cptr structure *does* not exist
					 * anymore!!! --msa
					 */
					return FLUSH_BUFFER;

				case ZIP_NEXT_BUFFER:
					if ((*ch2 == '\r' || *ch2 == '\n') && length) {
						ch2++;
						length--;
					}

					if (length) {
						int err;
						ch2 = zip_input(cptr->serv->zip_in, ch2, &length, &err, &nbuf, &nlen);

						if(length == -1) {
							sendto_realops("ZipIn error[2] for %s: (%d) %s\n", cptr->name, err, ch2);
							return exit_client(cptr, cptr, &me, "fatal error in compressed input");
						}
					}

					if (IsServer(cptr))
						sendto_syschan("links", "Incoming link from %s is now compressed.", cptr->name);
					break;
			}

			/* Socket is dead so exit (which always returns with
			 * FLUSH_BUFFER here).  - avalon
			 */
			if (cptr->flags & FLAGS_DEADSOCKET)
				return exit_client(cptr, cptr, &me, "Dead Socket");
			ch1 = cptr->buffer;
		}
		else if (ch1 < cptr->buffer + (sizeof(cptr->buffer)-1))
			ch1++; /* There is always room for the null */
	}
	cptr->count = ch1 - cptr->buffer;

	if (nbuf) {
		static time_t last_complain = 0;
		static int numrepeat = 0;

		numrepeat++;

		/* if more than 5 mins have elapsed */
		if(NOW > (last_complain + 300)) {
			if (last_complain == 0)
				sendto_realops("Overflowed zipInBuf! If you see this a lot, you should increase zipInBufSize in src/zlink.c.");
			else
				sendto_realops("Overflowed zipInBuf %d time%s in the last %d minutes. If you see this a lot, you should increase zipInBufSize in src/zlink.c.", numrepeat, numrepeat == 1 ? "" : "s", (NOW - last_complain) / 60);
			last_complain = NOW;
			numrepeat = 0;
		}

		buffer = nbuf;
		length = nlen;
		nbuf = NULL;
		goto zcontinue; /* gross, but it should work.. */
	}

	return 0;
}
