/* s_user.c -- user-related functions, almost unsorted;
 * Copyright (C) 2001-2002, ForestNet IRC Network
 * This is a free software; see file `license' for details.
 * $Id: s_user.c,v 1.62 2002/11/18 12:44:08 hex Exp $
 */

#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#ifndef _WIN32
# include <utmp.h>
#else
# include <io.h>
#endif

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "unicode.h"
#include "send.h"

void send_umode_out (aClient *, aClient *, int);
void send_umode (aClient *, aClient *, int, int, char *);
static is_silenced (aClient *, aClient *);

/* common services */
aClient *nsptr, *csptr, *msptr, *osptr;

static char buf[BUFSIZE], buf2[BUFSIZE];

/*
** m_functions execute protocol messages on this server:
**
**     cptr    is always NON-NULL, pointing to a *LOCAL* client
**             structure (with an open socket connected!). This
**             identifies the physical socket where the message
**             originated (or which caused the m_function to be
**             executed--some m_functions may call others...).
**
**     sptr    is the source of the message, defined by the
**             prefix part of the message if present. If not
**             or prefix not found, then sptr==cptr.
**
**             (!IsServer(cptr)) => (cptr == sptr), because
**             prefixes are taken *only* from servers...
**
**             (IsServer(cptr))
**                     (sptr == cptr) => the message didn't
**                     have the prefix.
**
**                     (sptr != cptr && IsServer(sptr) means
**                     the prefix specified servername. (?)
**
**                     (sptr != cptr && !IsServer(sptr) means
**                     that message originated from a remote
**                     user (not local).
**
**             combining
**
**             (!IsServer(sptr)) means that, sptr can safely
**             taken as defining the target structure of the
**             message in this server.
**
**     *Always* true (if 'parse' and others are working correct):
**
**     1)      sptr->from == cptr  (note: cptr->from == cptr)
**
**     2)      MyConnect(sptr) <=> sptr == cptr (e.g. sptr
**             *cannot* be a local connection, unless it's
**             actually cptr!). [MyConnect(x) should probably
**             be defined as (x == x->from) --msa ]
**
**     parc    number of variable parameter strings (if zero,
**             parv is allowed to be NULL)
**
**     parv    a NULL terminated list of parameter pointers,
**
**                     parv[0], sender (prefix string), if not present
**                             this points to an empty string.
**                     parv[1]...parv[parc-1]
**                             pointers to additional parameters
**                     parv[parc] == NULL, *always*
**
**             note:   it is guaranteed that parv[0]..parv[parc-1] are all
**                     non-NULL pointers.
*/

/*
** next_client
**     Local function to find the next matching client. The search
**     can be continued from the specified client entry. Normal
**     usage loop is:
**
**     for (x = client; x = next_client(x,mask); x = x->next)
**             HandleMatchingClient;
**
*/

aClient *next_client(aClient * next, char *ch)
{
	aClient *tmp = next;

	next = find_client(ch, tmp);

	if (tmp && tmp->prev == next)
		return NULL;

	if (next != tmp)
		return next;

	for (; next; next = next->next) {
		if (IsService(next))
			continue;
		if (!match(ch, next->name) || !match(next->name, ch))
			break;
	}

	return next;
}

/* hunt_server
 *
 *     Do the basic thing in delivering the message (command)
 *     across the relays to the specific server (server) for
 *     actions.
 *
 *     Note:   The command is a format string and *MUST* be
 *             of prefixed style (e.g. ":%s COMMAND %s ...").
 *             Command can have only max 8 parameters.
 *
 *     server  parv[server] is the parameter identifying the
 *             target server.
 *
 *     *WARNING*
 *             parv[server] is replaced with the pointer to the
 *             real servername from the matched client (I'm lazy
 *             now --msa).
 *
 *     returns: (see #defines)
 */
int hunt_server(aClient * cptr, aClient * sptr, char *command, int server, int parc, char *parv[])
{
	aClient *acptr;

	/* Assume it's me, if no server */
	if (parc <= server || BadPtr(parv[server])
	|| match(me.name, parv[server]) == 0 || match(parv[server], me.name) == 0)
		return (HUNTED_ISME);

	/* These are to pickup matches that would cause the following
	 * message to go in the wrong direction while doing quick fast
	 * non-matching lookups.
	 */
	if ((acptr = find_client(parv[server], NULL)))
		if (acptr->from == sptr->from && !MyConnect(acptr))
			acptr = NULL;

	if (!acptr && (acptr = find_server(parv[server], NULL)))
		if (acptr->from == sptr->from && !MyConnect(acptr))
			acptr = NULL;

	if (!acptr)
		for (acptr = client, (void) collapse(parv[server]); (acptr = next_client(acptr, parv[server])); acptr = acptr->next) {
			if (acptr->from == sptr->from && !MyConnect(acptr))
				continue;
			/* Fix to prevent looping in case the parameter for
			 * some reason happens to match someone from the from
			 * link --jto
			 */
			if (IsRegistered(acptr) && (acptr != cptr))
				break;
		}
	if (acptr) {
		if (IsMe(acptr) || MyClient(acptr))
			return HUNTED_ISME;
		if (match(acptr->name, parv[server]))
			parv[server] = acptr->name;
		sendto_one(acptr, command, parv[0], parv[1], parv[2], parv[3], parv[4], parv[5], parv[6], parv[7], parv[8]);
		return (HUNTED_PASS);
	}
	sendto_one(sptr, err_str(ERR_NOSUCHSERVER), me.name, parv[0], parv[server]);
	return (HUNTED_NOSUCH);
}


/* returns 0 if it's ok to send the message and 1 if it's not */
int check_for_target_limit(aClient *sptr, void *target, const char *name)
{
	time_t now;
	unsigned char hash, *src, *dst;
	unsigned int tmp = ((unsigned int)target & 0xffff00) >> 8;

	if (IsAnOper(sptr))
		return 0;

	hash = (tmp * tmp) >> 12;
	if (sptr->targets[0] == hash)
		return 0;

	/* check if this target already exists (ok if found) */
	for (dst = sptr->targets + MAXTARGETS - 1; dst != sptr->targets; --dst) {
		/* let the latest target always be the first in line */
		if (*dst == hash) {
			for (src = dst - 1; dst != sptr->targets; *dst-- = *src--);
			*dst = hash;
			return 0;
		}
	}

	if ((now = time(NULL)) < sptr->nexttarget) {
		if (sptr->nexttarget - now < TARGET_DELAY + 8) {
			sptr->nexttarget += 2;
			sendto_one(sptr, err_str(ERR_TARGETTOOFAST), me.name, sptr->name, name, sptr->nexttarget - now);
		}
		return 1;
	}

	sptr->nexttarget += TARGET_DELAY;

	if (sptr->nexttarget < now - (TARGET_DELAY * (MAXTARGETS - 1)))
			sptr->nexttarget = now - (TARGET_DELAY * (MAXTARGETS - 1));

	for (dst = sptr->targets + MAXTARGETS - 1, src = dst - 1; dst != sptr->targets; *dst-- = *src--);
	*dst = hash;

	return 0;
}

/* canonize
 *
 * reduce a string of duplicate list entries to contain only the unique
 * items.  Unavoidably O(n^2).
 */
char *canonize(char *buffer)
{
	static char cbuf[BUFSIZ];
	char *s, *t, *cp = cbuf, *p = NULL, *p2;
	int l = 0;

	*cp = '\0';

	for (s = strtoken(&p, buffer, ","); s; s = strtoken(&p, NULL, ",")) {
		if (l) {
			t = strtoken(&p2, cbuf, ",");
			for (p2 = NULL; t; t = strtoken(&p2, NULL, ",")) {
				if (!mycmp(s, t))
					break;
				else if (p2)
					p2[-1] = ',';
			}
		}
		else
			t = NULL;

		if (!t) {
			if (l)
				*(cp - 1) = ',';
			else
				l = 1;

			strcpy(cp, s);
			if (p)
				cp += (p - s);
		}
		else if (p2)
			p2[-1] = ',';
	}
	return cbuf;
}


void ircd_catch_service(aClient *acptr)
{
	char *who = NULL;

	if (nsptr == NULL && !strcasecmp(acptr->name, ns_nick)) {
		nsptr = acptr;
		who = "NickServ";
	}
	if (csptr == NULL && !strcasecmp(acptr->name, cs_nick)) {
		csptr = acptr;
		who = "ChanServ";
	}
	if (!msptr && !strcasecmp(acptr->name, ms_nick)) {
		msptr = acptr;
		who = "MemoServ";
	}
	if (!osptr && !strcasecmp(acptr->name, os_nick)) {
		osptr = acptr;
		who = "OperServ";
	}

	if (who != NULL)
		sendto_syschan("services", "found %s (%s)",
			who, acptr->name);
}


/*
** register_user
**     This function is called when both NICK and USER messages
**     have been accepted for the client, in whatever order. Only
**     after this the USER message is propagated.
**
**     NICK's must be propagated at once when received, although
**     it would be better to delay them too until full info is
**     available. Doing it is not so simple though, would have
**     to implement the following:
**
**     1) user telnets in and gives only "NICK foobar" and waits
**     2) another user far away logs in normally with the nick
**        "foobar" (quite legal, as this server didn't propagate
**        it).
**     3) now this server gets nick "foobar" from outside, but
**        has already the same defined locally. Current server
**        would just issue "KILL foobar" to clean out dups. But,
**        this is not fair. It should actually request another
**        nick from local user or kill him/her...
*/

int register_user(aClient *cptr, aClient *sptr, char *nick, char *username)
{
	aConfItem *aconf;
	char *parv[3], *tmpstr;

#ifdef HOSTILENAME
	char stripuser[USERLEN + 1], *u1 = stripuser, *u2, olduser[USERLEN + 1], userbad[USERLEN * 2 + 1], *ubad = userbad, noident = 0;
#endif
	short oldstatus = sptr->status, upper = 0, lower = 0, special = 0;
	anUser *user = sptr->user;
	int i;

	user->last = time(NULL);
	parv[0] = sptr->name;
	parv[1] = parv[2] = NULL;

	if (MyConnect(sptr)) {
		/* Borg88 Remove this 3 lines out if you want to have IDENT checking back */
		strncpyzt(user->username, username, USERLEN + 1);
		strncpyzt(sptr->username, username, USERLEN + 1);
		sptr->flags |= FLAGS_GOTID;

		sptr->flags |= FLAGS_PASSED;

		if ((i = check_client(sptr))) {
			if (i == -4)
				strcpy(gcomment, "Too many connections from your IP");
			else if (i == -3)
				strcpy(gcomment, "Too many connections");
			else
				strcpy(gcomment, "Unauthorized connection");

			sendto_syschan("clients", "Exit: %s (%s)", get_client_host(sptr), gcomment);

			ircstp->is_ref++;
			return exit_client(cptr, sptr, &me, gcomment);
		}
		/* added internal hosts support -- Borg88 */
		else if (*sptr->vhost)
			strncpyzt(user->host, sptr->vhost, HOSTLEN + 1);
		else if (sptr->hostp) {
			/* No control-chars or ip-like dns replies... I cheat :) --
			   OnyxDragon */
			for (tmpstr = sptr->sockhost; *tmpstr > ' ' && *tmpstr < 127; tmpstr++);
			if (*tmpstr || !*user->host || isdigit(*(tmpstr - 1)))
				/* Fix the sockhost for debug jic */
				strncpyzt(sptr->sockhost, (char *) inetntoa((char *) &sptr->ip), sizeof(sptr->sockhost));
			strncpyzt(user->host, sptr->sockhost, sizeof(sptr->sockhost));
		}
		else	/* Failsafe point, don't let the user define their own hostname
				   via the USER command --Cabal95 */
			strncpyzt(user->host, sptr->sockhost, HOSTLEN + 1);
		aconf = sptr->confs->value.aconf;
		/*
		 * I do not consider *, ~ or ! 'hostile' in usernames,
		 * as it is easy to differentiate them (Use \*, \? and \\)
		 * with the possible?
		 * exception of !. With mIRC etc. ident is easy to fake
		 * to contain @ though, so if that is found use non-ident
		 * username. -Donwulff
		 *
		 * I do, We only allow a-z A-Z 0-9 _ - and . now so the
		 * !strchr(sptr->username, '@') check is out of date. -Cabal95
		 *
		 * Moved the noident stuff here. -OnyxDragon
		 */
		/* Borg88 Code disabled and some of this stuff was copied up before
		   check_client() */
#if 0
		if (!(sptr->flags & FLAGS_DOID))
			strncpyzt(user->username, username, USERLEN + 1);
		else if (sptr->flags & FLAGS_GOTID)
			strncpyzt(user->username, sptr->username, USERLEN + 1);
		else {
			strncpyzt(user->username, username, USERLEN + 1);
			strncpyzt(sptr->username, username, USERLEN + 1);
			sptr->flags |= FLAGS_GOTID;
#ifdef HOSTILENAME
			noident = 1;
#endif
		}
#endif

#ifdef HOSTILENAME
		/*
		 * Limit usernames to just 0-9 a-z A-Z _ - and .
		 * It strips the "bad" chars out, and if nothing is left
		 * changes the username to the first 8 characters of their
		 * nickname. After the MOTD is displayed it sends numeric
		 * 455 to the user telling them what(if anything) happened.
		 * -Cabal95
		 *
		 * Moved the noident thing to the right place - see above
		 * -OnyxDragon
		 */
		for (u2 = user->username + noident; *u2; u2++) {
			if (isallowed(*u2))
				*u1++ = *u2;
			else if (*u2 < 32) {
				/*
				 * Make sure they can read what control
				 * characters were in their username.
				 */
				*ubad++ = '^';
				*ubad++ = *u2 + '@';
			}
			else
				*ubad++ = *u2;
		}
		*u1 = '\0';
		*ubad = '\0';
		if (strlen(stripuser) != strlen(user->username + noident)) {
			if (stripuser[0] == '\0') {
				strncpy(stripuser, cptr->name, 8);
				stripuser[8] = '\0';
			}

			strcpy(olduser, user->username + noident);
			strncpy(user->username + 1, stripuser, USERLEN - 1);
			user->username[0] = '~';
			user->username[USERLEN] = '\0';
		}
		else
			u1 = NULL;
#endif

		if (!BadPtr(aconf->passwd)) {
			if (!StrEq(sptr->passwd, aconf->passwd)) {
				ircstp->is_ref++;
				sendto_one(sptr, err_str(ERR_PASSWDMISMATCH), me.name, parv[0]);
				return exit_client(cptr, sptr, &me, "Bad Password");
			}
			/* .. Else password check was successful, clear the pass so it
			   doesn't get sent to NickServ. - Wizzu */
			else
				sptr->passwd[0] = '\0';
		}

		/* following block for the benefit of time-dependent K:-lines */
		if (find_kill(sptr)) {
			ircstp->is_ref++;
			sendto_syschan("clients", "Exit: %s (%s)", get_client_host(sptr), gcomment);
			return exit_client(cptr, sptr, &me, gcomment);
		}

#ifdef R_LINES
		if (find_restrict(sptr)) {
			ircstp->is_ref++;
			return exit_client(cptr, sptr, &me, "R-Lined");
		}
#endif

	/* server is closed */
	if (oper_closed)
		return exit_client(cptr, sptr, &me, "Server is closed by an administrator, please try again later.");

#ifdef DISALLOW_MIXED_CASE
/* check for mixed case usernames, meaning probably hacked -- Jon2 3-94 */
# ifdef IGNORE_CASE_FIRST_CHAR
		tmpstr = (username[0] == '~' ? &username[2] : &username[1]);
# else
		tmpstr = (username[0] == '~' ? &username[1] : username);
# endif /* IGNORE_CASE_FIRST_CHAR */
		while (*tmpstr && !(lower && upper || special)) {
			c = *tmpstr;
			tmpstr++;
			if (islower(c)) {
				lower++;
				continue;
			}
			if (isupper(c)) {
				upper++;
				continue;
			}
			if (c == '-' || c == '_' || c == '.' || isdigit(c))
				continue;
			special++;
		}
		if (lower && upper || special) {
			sendto_ops("Invalid username: %s", get_client_name(sptr, FALSE));
			ircstp->is_ref++;
			return exit_client(cptr, sptr, sptr, "Invalid username");
		}

#endif /* DISALLOW_MIXED_CASE */

		if (oldstatus == STAT_MASTER && MyConnect(sptr))
			m_oper(&me, sptr, 1, parv);
	}
	else strncpyzt(user->username, username, USERLEN + 1);

	SetClient(sptr);
	hmask_update(sptr);

	if (MyConnect(sptr)) {
		/* attach the desired codepage and translate realname,
		 * since it was accepted as unicode.
		 */
		if ((sptr->charset = charset_attach(
		sptr->acpt->confs->value.aconf->name)) != NULL)
			charset_convertto(sptr->info, sizeof(sptr->info),
				sptr->charset);

		sendto_one(sptr, rpl_str(RPL_WELCOME), me.name, nick,
			network_name ? network_name : "the IRC Network", nick);
		/* This is a duplicate of the NOTICE but see below... */
		sendto_one(sptr, rpl_str(RPL_YOURHOST), me.name, nick, me.name);
		sendto_one(sptr, rpl_str(RPL_CREATED), me.name, nick, creation);
		sendto_one(sptr, rpl_str(RPL_MYINFO), me.name, parv[0], me.name,
			get_ircd_version());
		send_protoctl(sptr);

		if (force_motd)
			m_motd(sptr, sptr, 1, parv);
		else
			sendto_one(sptr, rpl_str(RPL_ENDOFMOTD),
				me.name, parv[0]);

		m_lusers(sptr, sptr, 1, parv);
#ifdef HOSTILENAME
		/* Now send a numeric to the user telling them what, if anything, happened. */
		if (u1)
			sendto_one(sptr, err_str(ERR_HOSTILENAME), me.name, sptr->name, olduser, userbad, stripuser);
#endif
		nextping = time(NULL);
		sendto_syschan("clients", "Connected: %s[%s@%s], fd = %d", nick, user->username, user->host, sptr->fd);
	}
	else if (IsServer(cptr)) {
		aClient *acptr;

		if (!(acptr = sptr->srvptr)) {
			sendto_ops("Bad USER [%s] :%s USER %s : No such server", cptr->name, nick, user->username);
			sendto_one(cptr, ":%s KILL %s :%s (No such server)", me.name, sptr->name, me.name);
			sptr->flags |= FLAGS_KILLED;
			return exit_client(sptr, sptr, &me, "USER without prefix(2.8) or wrong prefix");
		}
		else if (acptr->from != sptr->from) {
			sendto_ops("Bad User [%s] :%s USER %s %s, != %s[%s]", cptr->name, nick, user->username, acptr->srvptr->name, acptr->name, acptr->from->name);
			sendto_one(cptr, ":%s KILL %s :%s (%s != %s[%s])", me.name, sptr->name, me.name, acptr->srvptr->name, acptr->from->name, acptr->from->sockhost);
			sptr->flags |= FLAGS_KILLED;
			return exit_client(sptr, sptr, &me, "USER server wrong direction");
		}
		else
			sptr->flags |= (acptr->flags & FLAGS_TS8);

		/* *FINALL* this gets in ircd... -- Barubary  */
		if (find_conf_host(cptr->confs, sptr->name, CONF_UWORLD) || (sptr->user && find_conf_host(cptr->confs, sptr->srvptr->name, CONF_UWORLD))) {
			sptr->flags |= FLAGS_ULINE;
			if (!BadPtr(services_host) && !MyClient(sptr))
				ircd_catch_service(sptr);
		}
	}

	hash_check_notify(sptr, RPL_LOGON); /* Uglier hack */
	sendto_serv_butone(cptr, "NICK %s %d %d %s %s %s %lu :%s", nick, sptr->hopcount + 1, sptr->lastnick, user->username, user->host, sptr->srvptr->name, sptr->user->servicestamp, sptr->info);

	/* Send password from sptr->passwd to NickServ for identification, if
	   passwd given and if NickServ is online. - by taz, modified by Wizzu */
	if (MyConnect(sptr)) {
		send_umode_out(cptr, sptr, 0);
		if (sptr->passwd[0] && nsptr)
			sendto_one(nsptr, ":%s PRIVMSG %s@%s :IDENTIFY %s", sptr->name, ns_nick, services_host, sptr->passwd);
	}

	if (MyConnect(sptr) && !BadPtr(sptr->passwd))
		memset(sptr->passwd, 0, sizeof(sptr->passwd));

	return 0;
}


/*
** m_message (used in m_private() and m_notice())
** the general function to deliver MSG's between users/channels
**
**     parv[0] = sender prefix
**     parv[1] = receiver list
**     parv[2] = message text
**
** massive cleanup
** rev argv 6/91
**
*/

static int m_message(aClient *cptr, aClient *sptr, int parc, char *parv[], int notice)
{
	aClient *acptr;
	aChannel *chptr;
	char *s, *nick, *server, *p, *cmd;

	if (notice ? check_registered(sptr) : check_registered_user(sptr))
		return 0;

	sptr->flags &= ~FLAGS_TS8;

	cmd = notice ? MSG_NOTICE : MSG_PRIVATE;

	if (parc < 2 || *parv[1] == '\0') {
		sendto_one(sptr, err_str(ERR_NORECIPIENT), me.name, parv[0], cmd);
		return -1;
	}

	if (parc < 3 || *parv[2] == '\0') {
		sendto_one(sptr, err_str(ERR_NOTEXTTOSEND), me.name, parv[0]);
		return -1;
	}

	if (MyConnect(sptr))
		parv[1] = canonize(parv[1]);

	for (p = NULL, nick = strtoken(&p, parv[1], ","); nick; nick = strtoken(&p, NULL, ",")) {
		if ((server = (char*)strchr(nick, '@')) != NULL && (strcasecmp(server+1, me.name) == 0 || server[1] == '\0'))
			*server = '\0';

		/* nickname addressed? */
		if ((acptr = find_person(nick, NULL))) {
			if (MyClient(sptr) && check_for_target_limit(sptr, acptr, acptr->name))
				continue;

			if (!is_silenced(sptr, acptr)) {
				if (!notice && IsAway(acptr) && MyClient(sptr))
					sendto_one(sptr, rpl_str(RPL_AWAY), me.name, parv[0], acptr->name, MyClient(acptr) ? acptr->away : "Gone.");
				sendto_prefix_one(acptr, sptr, ":%s %s %s :%s", parv[0], cmd, nick, parv[2]);
			}
			continue;
		}

		if (nick[0] == '@') {
			/* If its a message for all Channel OPs call the function
			   sendto_channelops_butone() to handle it.  -Cabal95 */
			if (nick[1] == '#') {
				if (chptr = find_channel(nick + 1, NullChn)) {
					if (can_send(sptr, chptr) == 0 || IsULine(cptr, sptr)) {
						if (ColorBlindChannel(chptr))
							strip_color(parv[2]);
						sendto_channelops_butone(cptr, sptr, chptr, ":%s %s %s :%s", parv[0], cmd, nick, parv[2]);
					}
					else if (!notice)
						sendto_one(sptr, err_str(ERR_CANNOTSENDTOCHAN), me.name, parv[0], nick + 1);
				}
				else
					sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, parv[0], nick + 1);
			}

			/* If its a message for all Channel ops and voices call the
			   function sendto_channelvoice_butone() to handle it.  -DuffJ */
			else if (nick[1] == '+' && nick[2] == '#') {
				if (chptr = find_channel(nick + 2, NullChn)) {
					if (can_send(sptr, chptr) == 0 || IsULine(cptr, sptr)) {
						if (ColorBlindChannel(chptr))
							strip_color(parv[2]);
						sendto_channelvoice_butone(cptr, sptr, chptr, ":%s %s %s :%s", parv[0], cmd, nick, parv[2]);
					}
					else if (!notice)
						sendto_one(sptr, err_str(ERR_CANNOTSENDTOCHAN), me.name, parv[0], nick + 2);
				}
				else
					sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, parv[0], nick + 2);
			}
			continue;
		}

		/* Now allows U-lined users to send to channel no problemo -- Barubary */
		if ((chptr = find_channel(nick, NullChn))) {
			if (can_send(sptr, chptr) == 0) {
				if (MyClient(sptr) && check_for_target_limit(sptr, chptr, chptr->chname))
					continue;
				if (ColorBlindChannel(chptr))
					strip_color(parv[2]);
				sendto_channel_butone(cptr, sptr, chptr, ":%s %s %s :%s", parv[0], cmd, nick, parv[2]);
			}
			else if (!notice)
				sendto_one(sptr, err_str(ERR_CANNOTSENDTOCHAN), me.name, parv[0], nick);
			continue;
		}

		/* the following two cases allow masks in NOTICEs (for OPERs only)
		   Armin, 8Jun90 (gruner@informatik.tu-muenchen.de) */
		if ((*nick == '$' || *nick == '#') && IsAnOper(sptr)) {
			if (!(s = (char *) strrchr(nick, '.'))) {
				sendto_one(sptr, err_str(ERR_NOTOPLEVEL), me.name, parv[0], nick);
				continue;
			}
			while (*++s)
				if (*s == '.' || *s == '*' || *s == '?')
					break;
			if (*s == '*' || *s == '?') {
				sendto_one(sptr, err_str(ERR_WILDTOPLEVEL), me.name, parv[0], nick);
				continue;
			}
			sendto_match_butone(IsServer(cptr) ? cptr : NULL, sptr, nick + 1, (*nick == '#') ? MATCH_HOST : MATCH_SERVER, ":%s %s %s :%s", parv[0], cmd, nick, parv[2]);
			continue;
		}

		if (server++ != NULL) {
			if ((acptr = find_server(server, NULL)) != NULL) {
				sendto_one(acptr, ":%s %s %s :%s", parv[0], cmd, nick, parv[2]);
				continue;
			}

			if (!BadPtr(services_host) && strcasecmp(server, services_host) == 0) {
				sendto_one(sptr, err_str(ERR_SERVICESDOWN), me.name, parv[0], nick);
				continue;
			}

			sendto_one(sptr, err_str(ERR_NOSUCHSERVER), me.name, parv[0], server);
			continue;
		}

		sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, parv[0], nick);
	}

	return 0;
}

/* parv[0] = sender prefix, parv[1] = receiver list, parv[2] = message text */
int m_private(aClient * cptr, aClient * sptr, int parc, char *parv[])
{
	return m_message(cptr, sptr, parc, parv, 0);
}

/* parv[0] = sender prefix, parv[1] = receiver list, parv[2] = notice text */
int m_notice(aClient * cptr, aClient * sptr, int parc, char *parv[])
{
	return m_message(cptr, sptr, parc, parv, 1);
}


/***************************************************************************
 * m_pass() - Added Sat, 4 March 1989
 ***************************************************************************/

/*
** m_pass
**     parv[0] = sender prefix
**     parv[1] = password
*/
int m_pass(aClient * cptr, aClient * sptr, int parc, char *parv[])
{
	char *password = parc > 1 ? parv[1] : NULL;

	if (BadPtr(password)) {
		sendto_one(cptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "PASS");
		return 0;
	}

	if (!MyConnect(sptr) || (!IsUnknown(cptr) && !IsHandshake(cptr))) {
		sendto_one(cptr, err_str(ERR_ALREADYREGISTRED), me.name, parv[0]);
		return 0;
	}

	strncpyzt(cptr->passwd, password, sizeof(cptr->passwd));
	return 0;
}

/*
 * m_ison added by Darren Reed 13/8/91 to act as an efficent user indicator
 * with respect to cpu/bandwidth used. Implemented for NOTIFY feature in
 * clients. Designed to reduce number of whois requests. Can process
 * nicknames in batches as long as the maximum buffer length.
 *
 * format:
 * ISON :nicklist
 */

int m_ison(aClient * cptr, aClient * sptr, int parc, char *parv[])
{
	char namebuf[USERLEN + HOSTLEN + 4];
	aClient *acptr;
	char *s, **pav = parv, *user;
	int len;
	char *p = NULL;

	if (check_registered(sptr))
		return 0;

	if (parc < 2) {
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "ISON");
		return 0;
	}

	(void) sprintf(buf, rpl_str(RPL_ISON), me.name, *parv);
	len = strlen(buf);

	for (s = strtoken(&p, *++pav, " "); s; s = strtoken(&p, NULL, " ")) {
		if (user = strchr(s, '!'))
			*user++ = '\0';
		if ((acptr = find_person(s, NULL))) {
			if (user) {
				strcpy(namebuf, acptr->user->username);
				strcat(namebuf, "@");
				strcat(namebuf, acptr->user->host);
				if (match(user, namebuf))
					continue;
				*--user = '!';
			}

			(void) strncat(buf, s, sizeof(buf) - len);
			len += strlen(s);
			(void) strncat(buf, " ", sizeof(buf) - len);
			len++;
		}
	}
	sendto_one(sptr, "%s", buf);
	return 0;
}

static int user_modes[] = {
	UMODE_OPER, 'o',
	UMODE_LOCOP, 'O',
	UMODE_INVISIBLE, 'i',
	UMODE_WALLOP, 'w',
	UMODE_FAILOP, 'g',
	UMODE_SERVNOTICE, 's',
	UMODE_KILLS, 'k',
	UMODE_SADMIN, 'a',
	UMODE_ADMIN, 'A',
	UMODE_FLOOD, 'f',
	UMODE_REGNICK, 'r',
	UMODE_HMASK, 'x',
	UMODE_GONE, 'G',
	UMODE_UWNOTICE, 'n',
	0, 0
};

/* m_umode() added 15/10/91 By Darren Reed.
 * parv[0] - sender
 * parv[1] - username to change mode for
 * parv[2] - modes to change
 */
int m_umode(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	int flag, *s, what, setflags;
	char **p, *m;
	aClient *acptr;

	if (check_registered_user(sptr))
		return 0;

	what = MODE_ADD;

	if (parc < 2) {
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "MODE");
		return 0;
	}

	if (!(acptr = find_person(parv[1], NULL))) {
		if (MyConnect(sptr))
			sendto_one(sptr, err_str(ERR_NOSUCHCHANNEL), me.name, parv[0], parv[1]);
		return 0;
	}

	if (IsServer(sptr) || (sptr != acptr && (!IsOper(sptr) || parc > 2))) {
		if (IsServer(sptr))
			sendto_ops_butone(NULL, &me, ":%s WALLOPS :MODE for User %s From %s!%s", me.name, parv[1], get_client_name(cptr, FALSE), sptr->name);
		else
			sendto_one(sptr, err_str(ERR_USERSDONTMATCH), me.name, parv[0]);
		return 0;
	}

	if (parc < 3) {
		m = buf;
		*m++ = '+';
		for (s = user_modes; (flag = *s) && (m - buf < BUFSIZE - 4); s += 2)
			if ((acptr->umodes & flag))
				*m++ = (char) (*(s + 1));
		*m = '\0';
		sendto_one(sptr, rpl_str(RPL_UMODEIS), me.name, parv[1], buf);
		return 0;
	}

	/* find flags already set for user */
	setflags = sptr->umodes;

	/* parse mode change string(s) */
	for (p = &parv[2]; p && *p; p++) {
		for (m = *p; *m; m++) {
			switch (*m) {
			case '+':
				what = MODE_ADD;
				break;
			case '-':
				what = MODE_DEL;
				break;
				/* we may not get these, but they shouldnt
				 * be in default
				 */
			case ' ':
			case '\n':
			case '\r':
			case '\t':
				break;

			case 'G':
				if (!IsServer(cptr))
					break;

			default:
				if (*m == 'r' && !IsServer(cptr)) {
					sendto_one(sptr, err_str(ERR_UMODEUNKNOWNFLAG), me.name, parv[0], *m);
					break;
				}

				for (s = user_modes; (flag = *s); s += 2)
					if (*m == (char) (*(s + 1))) {
						if (what == MODE_ADD)
							sptr->umodes |= flag;
						else
							sptr->umodes &= ~flag;
						break;
					}
				if (flag == 0 && MyConnect(sptr))
					sendto_one(sptr, err_str(ERR_UMODEUNKNOWNFLAG), me.name, parv[0], *m);
				break;
			}

			if (*m == 'x')
				hmask_update(sptr);
		}
	}

	/* stop users making themselves operators too easily */
	if (!IsServer(cptr)) {
		if (!(setflags & UMODE_OPER) && IsOper(sptr))
			ClearOper(sptr);
		if (!(setflags & UMODE_LOCOP) && IsLocOp(sptr))
			sptr->umodes &= ~UMODE_LOCOP;
		if (!(setflags & UMODE_WALLOP))
			sptr->umodes &= ~UMODE_WALLOP;

		if (!IsAnOper(sptr))
			sptr->umodes &= ~(UMODE_FAILOP | UMODE_SERVNOTICE | UMODE_KILLS | /*UMODE_FLOOD |*/ UMODE_ADMIN | UMODE_SADMIN);
	}

	/* New oper access flags - Only let them set certian usermodes on
	 * themselves IF they have access to set that specific mode in their
	 * O:Line.
	 */
	if (MyClient(sptr) && IsAnOper(sptr)) {
		if (IsAdmin(sptr) && !OPIsAdmin(sptr))
			ClearAdmin(sptr);
		if (IsSAdmin(sptr) && !OPIsSAdmin(sptr))
			ClearSAdmin(sptr);
	}

	/* If I understand what this code is doing correctly...
	 * If the user WAS an operator and has now set themselves -o/-O
	 * then remove their access, d'oh!
	 * In order to allow opers to do stuff like go +o, +h, -o and
	 * remain +h, I moved this code below those checks. It should be
	 * O.K. The above code just does normal access flag checks. This
	 * only changes the operflag access level.  -Cabal95
	 */
	if ((setflags & (UMODE_OPER | UMODE_LOCOP)) && !IsAnOper(sptr) && MyConnect(sptr)) {
		det_confs_butmask(sptr, CONF_CLIENT & ~CONF_OPS);
		sptr->oflag = 0;
	}

	/* compare new flags with old flags and send string which
	 * will cause servers to update correctly.
	 */
	send_umode_out(cptr, sptr, setflags);

	hmask_update(sptr);

	return 0;
}

/* m_svsmode() added by taz
 * parv[0] - sender
 * parv[1] - username to change mode for
 * parv[2] - modes to change
 * parv[3] - Service Stamp (if mode == d)
 */
int m_svsmode(aClient * cptr, aClient * sptr, int parc, char *parv[])
{
	int flag, what, setflags, *s;
	char **p, *m;
	aClient *acptr;

	if (!IsULine(cptr, sptr))
		return 0;

	what = MODE_ADD;

	if (parc < 3)
		return 0;

	if (!(acptr = find_person(parv[1], NULL)))
		return 0;

	setflags = 0;
	for (s = user_modes; (flag = *s); s += 2)
		if (acptr->umodes & flag)
			setflags |= flag;
	/* parse mode change string(s) */
	for (p = &parv[2]; p && *p; p++)
		for (m = *p; *m; m++)
			switch (*m) {
			case '+':
				what = MODE_ADD;
				break;
			case '-':
				what = MODE_DEL;
				break;
				/* we may not get these, but they shouldnt be in default */
			case ' ':
			case '\n':
			case '\r':
			case '\t':
				break;
			case 'l':
				sendto_syschan("debug", ":%s SVSMODE %s %s %s", parv[0], parv[1], parv[2], parc > 2 ? parv[3] : "(null)");
				/* disabled this until we fix the services -- hex, 02-04-2002
				if (parv[3] && isdigit(*parv[3]))
					lu_mglobalu = atoi(parv[3]);
				*/
				break;
			case 'd':
				if (parv[3] && isdigit(*parv[3]))
					acptr->user->servicestamp = atol(parv[3]);
				break;
			default:
				for (s = user_modes; (flag = *s); s += 2)
					if (*m == (char) (*(s + 1))) {
						if (what == MODE_ADD)
							acptr->umodes |= flag;
						else
							acptr->umodes &= ~flag;
						break;
					}
				break;
			}

	if (parc > 3)
		sendto_serv_butone(cptr, ":%s SVSMODE %s %s %s", parv[0], parv[1], parv[2], parv[3]);
	else
		sendto_serv_butone(cptr, ":%s SVSMODE %s %s", parv[0], parv[1], parv[2]);

	if (MyClient(acptr))
		send_umode(acptr, acptr, setflags, ALL_UMODES, buf);

	return 0;
}

/*
 * send the MODE string for user (user) to connection cptr
 * -avalon
 */
void send_umode(aClient *cptr, aClient *sptr, int old, int sendmask, char *umode_buf)
{
	int *s, flag, what = MODE_NULL;
	char *m;

	/* prevent clients from seeing internal AWAY flag */
	if (cptr && IsClient(cptr))
		sendmask &= ~UMODE_GONE;

	/* build a string in umode_buf to represent the change in the user's
	 * mode between the new (sptr->flag) and 'old'.
	 */
	m = umode_buf;
	*m = '\0';
	for (s = user_modes; (flag = *s); s += 2) {
		if (MyClient(sptr) && !(flag & sendmask))
			continue;
		if ((flag & old) && !(sptr->umodes & flag)) {
			if (what == MODE_DEL)
				*m++ = *(s + 1);
			else {
				what = MODE_DEL;
				*m++ = '-';
				*m++ = *(s + 1);
			}
		}
		else if (!(flag & old) && (sptr->umodes & flag)) {
			if (what == MODE_ADD)
				*m++ = *(s + 1);
			else {
				what = MODE_ADD;
				*m++ = '+';
				*m++ = *(s + 1);
			}
		}
	}

	*m = '\0';
	if (*umode_buf && cptr)
		sendto_one(cptr, ":%s %s %s :%s", sptr->name, (IsToken(cptr) ? TOK_MODE : MSG_MODE), sptr->name, umode_buf);
}

/* added Sat Jul 25 07:30:42 EST 1992 */
void send_umode_out(aClient * cptr, aClient * sptr, int old)
{
	int i;
	aClient *acptr;

	send_umode(NULL, sptr, old, SEND_UMODES, buf);

#ifdef NPATH
	check_command((long) 4, ":%s MODE %s :%s", sptr->name, sptr->name, buf);
#endif

	for (i = highest_fd; i >= 0; i--)
		if ((acptr = local[i]) && IsServer(acptr) && (acptr != cptr) && (acptr != sptr) && *buf)
			sendto_one(acptr, ":%s MODE %s :%s", sptr->name, sptr->name, buf);

	if (cptr && MyClient(cptr))
		send_umode(cptr, sptr, old, ALL_UMODES, buf);
}

/***********************************************************************
 * m_silence() - Added 19 May 1994 by Run.
 *
 ***********************************************************************/

/* is_silenced : Does the actual check wether sptr is allowed
 *               to send a message to acptr.
 *               Both must be registered persons.
 * If sptr is silenced by acptr, his message should not be propagated,
 * but more over, if this is detected on a server not local to sptr
 * the SILENCE mask is sent upstream.
 */
static int is_silenced(aClient * sptr, aClient * acptr)
{
	struct SLink *lp;
	struct User *user;
	static char sender[HOSTLEN + NICKLEN + USERLEN + 5];

	if (!(acptr->user) || !(lp = acptr->user->silence) || !(user = sptr->user))
		return 0;

	sprintf(sender, "%s!%s@%s", sptr->name, user->username, user->host);

	for (; lp; lp = lp->next) {
		if (!match(lp->value.cp, sender)) {
			if (!MyConnect(sptr)) {
				sendto_one(sptr->from, ":%s SILENCE %s :%s", acptr->name, sptr->name, lp->value.cp);
				lp->flags = 1;
			}
			return 1;
		}
	}

	return 0;
}

int del_silence(aClient * sptr, char *mask)
{
	struct SLink **lp, *tmp;

	for (lp = &(sptr->user->silence); *lp; lp = &((*lp)->next))
		if (mycmp(mask, (*lp)->value.cp) == 0) {
			tmp = *lp;
			*lp = tmp->next;
			MyFree(tmp->value.cp);
			free_link(tmp);
			return 0;
		}
	return -1;
}

static int add_silence(aClient * sptr, char *mask)
{
	struct SLink *lp;
	int cnt = 0, len = 0;

	for (lp = sptr->user->silence; lp; lp = lp->next) {
		len += strlen(lp->value.cp);
		if (MyClient(sptr))
			if ((len > MAXSILELENGTH) || (++cnt >= MAXSILES)) {
				sendto_one(sptr, err_str(ERR_SILELISTFULL), me.name, sptr->name, mask);
				return -1;
			}
			else {
				if (!match(lp->value.cp, mask))
					return -1;
			}
		else if (!mycmp(lp->value.cp, mask))
			return -1;
	}

	lp = make_link();
	memset(lp, 0, sizeof(Link));
	lp->next = sptr->user->silence;
	lp->value.cp = (char *) MyMalloc(strlen(mask) + 1);
	(void) strcpy(lp->value.cp, mask);
	sptr->user->silence = lp;
	return 0;
}

/*
** m_silence
**     parv[0] = sender prefix
** From local client:
**     parv[1] = mask (NULL sends the list)
** From remote client:
**     parv[1] = nick that must be silenced
**      parv[2] = mask
*/
int m_silence(aClient * cptr, aClient * sptr, int parc, char *parv[])
{
	struct SLink *lp;
	aClient *acptr;
	char c, *cp;

	if (check_registered_user(sptr))
		return 0;

	if (MyClient(sptr)) {
		acptr = sptr;
		if (parc < 2 || *parv[1] == '\0' || (acptr = find_person(parv[1], NULL))) {
			if (!(acptr->user))
				return 0;
			for (lp = acptr->user->silence; lp; lp = lp->next)
				sendto_one(sptr, rpl_str(RPL_SILELIST), me.name, sptr->name, acptr->name, lp->value.cp);
			sendto_one(sptr, rpl_str(RPL_ENDOFSILELIST), me.name, acptr->name);
			return 0;
		}
		cp = parv[1];
		c = *cp;
		if (c == '-' || c == '+')
			cp++;
		else if (!(strchr(cp, '@') || strchr(cp, '.') || strchr(cp, '!') || strchr(cp, '*'))) {
			sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, parv[0], parv[1]);
			return -1;
		}
		else
			c = '+';
		cp = pretty_mask(cp);
		if ((c == '-' && !del_silence(sptr, cp)) || (c != '-' && !add_silence(sptr, cp))) {
			sendto_prefix_one(sptr, sptr, ":%s SILENCE %c%s", parv[0], c, cp);
			if (c == '-')
				sendto_serv_butone(NULL, ":%s SILENCE * -%s", sptr->name, cp);
		}
	}
	else if (parc < 3 || *parv[2] == '\0') {
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "SILENCE");
		return -1;
	}
	else if ((c = *parv[2]) == '-' || (acptr = find_person(parv[1], NULL))) {
		if (c == '-') {
			if (!del_silence(sptr, parv[2] + 1))
				sendto_serv_butone(cptr, ":%s SILENCE %s :%s", parv[0], parv[1], parv[2]);
		}
		else {
			(void) add_silence(sptr, parv[2]);
			if (!MyClient(acptr))
				sendto_one(acptr, ":%s SILENCE %s :%s", parv[0], parv[1], parv[2]);
		}
	}
	else {
		sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, parv[0], parv[1]);
		return -1;
	}
	return 0;
}
