/* win32.c -- nt service support, etc;
 * Copyright (C) 2001-2002 ForestNet IRC Network;
 * This is a free software; see file `license' for details.
 * $Id: win32.c,v 1.8 2002/11/18 12:44:09 hex Exp $
 */

#include <stdlib.h>
#include <stdio.h>
#include <windows.h>

#include "struct.h" /* argh, we need me.name */
#include "h.h"
#include "ircdlog.h"

extern int portnum;
extern char *configfile, *forced_server_name;
extern aClient me;

static SERVICE_STATUS svc_status;
static SERVICE_STATUS_HANDLE svc_handle;
static DWORD svc_error;
static HANDLE hEventSource = NULL;
static char *service_name = NULL;

extern void child_monitor(void);

static void print_last_error(void)
{
	char tmpbuf[1024];

	FormatMessage(
		FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
		NULL,
		GetLastError() & 0xFFFF,
		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
		tmpbuf,
		sizeof(tmpbuf),
		NULL);

	tmpbuf[sizeof(tmpbuf) - 1] = '\0';
	fprintf(stdout, "ERROR: %s", tmpbuf);
}


void vsyslog(int priority, const char *format, va_list vl)
{
	WORD type;
	char tmpbuf[1024], header[1024];
	const char *args[9];

	if (hEventSource == NULL)
		return;

	switch (priority) {
		case LOG_EMERG:
		case LOG_ALERT:
		case LOG_CRIT:
		case LOG_ERR:
			type = EVENTLOG_ERROR_TYPE;
			break;
		case LOG_WARNING:
			type = EVENTLOG_WARNING_TYPE;
			break;
		default:
			type = EVENTLOG_INFORMATION_TYPE;
	}

	sprintf(header, "Faerion IRC Daemon (service name: %s) reported the "
		"following:\r\n", service_name);

	args[0] = header;
	args[1] = tmpbuf;
	args[2] = "\r\n";
	args[3] = NULL;
	args[4] = NULL;
	args[5] = NULL;
	args[6] = NULL;
	args[7] = NULL;
	args[8] = NULL;

	/* getting around the spaces */
	tmpbuf[0] = '\r';
	tmpbuf[1] = '\n';

	vsnprintf(tmpbuf + 2, sizeof(tmpbuf) - 2, format, vl);

	tmpbuf[sizeof(tmpbuf)-1] = '\0';
	ReportEvent(hEventSource, type, 0, 3299, NULL, 9, 0, args, NULL);
}


void syslog(int priority, const char *format, ...)
{
	va_list vl;
	va_start(vl, format);
	vsyslog(priority, format, vl);
	va_end(vl);
}


static long registry_store(char *service_name)
{
	long ret;
	HKEY hKey;
	char key_name[1024], tmpbuf[1024];

	sprintf(key_name, "System\\CurrentControlSet\\Services\\%s\\Parameters",
		service_name);

	if ((ret = RegCreateKeyEx(HKEY_LOCAL_MACHINE, key_name, 0, NULL,
	REG_OPTION_NON_VOLATILE, KEY_WRITE, NULL, &hKey, NULL))
	!= ERROR_SUCCESS)
		return ret;

	if (stricmp(configfile, "ircd.conf"))
		RegSetValueEx(hKey, "ConfigurationFile", 0, REG_SZ,
			configfile, strlen(configfile));
	else
		RegDeleteValue(hKey, "ConfigurationFile");

	if (portnum != -1)
		RegSetValueEx(hKey, "ServerPort", 0, REG_DWORD,
			(LPBYTE)&portnum, sizeof(portnum));
	else
		RegDeleteValue(hKey, "ServerPort");

	if (log_verbosity != 0)
		RegSetValueEx(hKey, "LogVerbosity", 0, REG_DWORD,
			(LPBYTE)&log_verbosity, sizeof(log_verbosity));
	else
		RegDeleteValue(hKey, "LogVerbosity");

	if (forced_server_name != NULL)
		RegSetValueEx(hKey, "ServerName", 0, REG_SZ,
			forced_server_name, strlen(forced_server_name));
	else
		RegDeleteValue(hKey, "ServerName");

	if (GetCurrentDirectory(sizeof(tmpbuf), tmpbuf) != 0) {
		tmpbuf[sizeof(tmpbuf) - 1] = '\0';
		RegSetValueEx(hKey, "WorkingDirectory", 0, REG_SZ,
			tmpbuf, strlen(tmpbuf));
	}
	else
		RegDeleteValue(hKey, "WorkingDirectory");

	return RegCloseKey(hKey);
}


static long registry_restore(char *service_name)
{
	long ret;
	int reg_port;
	HKEY hKey;
	DWORD size, type;
	char key_name[1024], tmpbuf[1024];

	/* refresh information needed to use EventLog */

	if ((ret = RegCreateKeyEx(HKEY_LOCAL_MACHINE, "System"
	"\\CurrentControlSet\\Services\\EventLog\\Application"
	"\\Faerion IRC Daemon", 0, NULL, REG_OPTION_NON_VOLATILE,
	KEY_WRITE, NULL, &hKey, NULL)) != ERROR_SUCCESS)
		return ret;

	strcpy(key_name, "%SystemRoot%\\System32\\netmsg.dll");

	RegSetValueEx(hKey, "EventMessageFile", 0, REG_EXPAND_SZ,
		(LPBYTE)key_name, strlen(key_name));

	type = EVENTLOG_ERROR_TYPE | EVENTLOG_WARNING_TYPE |
		EVENTLOG_INFORMATION_TYPE;

	RegSetValueEx(hKey, "TypesSupported", 0, REG_DWORD,
		(LPBYTE)&type, sizeof(type));

	RegCloseKey(hKey);

	/* restoring the parameters */

	sprintf(key_name, "System\\CurrentControlSet\\Services\\%s\\Parameters",
		service_name);

	if ((ret = RegOpenKeyEx(HKEY_LOCAL_MACHINE, key_name, 0,
	KEY_QUERY_VALUE, &hKey)) != ERROR_SUCCESS)
		return ret;

	size = sizeof(tmpbuf) - 1;
	if (RegQueryValueEx(hKey, "ConfigurationFile", 0, &type,
	(LPBYTE)tmpbuf, &size) == ERROR_SUCCESS && type == REG_SZ) {
		tmpbuf[size] = '\0';
		/* we don't ever free this, but since it's a one-time allocation
		 * that is never changed and always used, we leave it as it is.
		 */
		configfile = strdup(tmpbuf);
	}

	size = sizeof(tmpbuf) - 1;
	if (RegQueryValueEx(hKey, "ServerName", 0, &type, (LPBYTE)tmpbuf,
	&size) == ERROR_SUCCESS && type == REG_SZ) {
		tmpbuf[size] = '\0';
		forced_server_name = strdup(tmpbuf);
	}

	size = sizeof(reg_port);
	if (RegQueryValueEx(hKey, "ServerPort", 0, &type, (LPBYTE)&reg_port,
	&size) == ERROR_SUCCESS && type == REG_DWORD)
		portnum = reg_port;

	size = sizeof(tmpbuf) - 1;
	if (RegQueryValueEx(hKey, "WorkingDirectory", 0, &type, (LPBYTE)tmpbuf,
	&size) == ERROR_SUCCESS && type == REG_SZ) {
		tmpbuf[size] = '\0';
		if (SetCurrentDirectory(tmpbuf) == 0) {
			RegCloseKey(hKey);
			return ERROR_FILE_NOT_FOUND;
		}
	}

	return RegCloseKey(hKey);
}


/* set current status of the service and report it to the manager */
static BOOL ReportStatusToSCMgr(DWORD curstate, DWORD exitcode, DWORD waithint)
{
	static DWORD checkpoint = 1;

	if (curstate == SERVICE_START_PENDING)
		svc_status.dwControlsAccepted = 0;
	else
		svc_status.dwControlsAccepted = SERVICE_ACCEPT_STOP;

	svc_status.dwCurrentState = curstate;
	svc_status.dwWin32ExitCode = exitcode;
	svc_status.dwWaitHint = waithint;

	if ((curstate == SERVICE_RUNNING) || (curstate == SERVICE_STOPPED))
		svc_status.dwCheckPoint = 0;
	else
		svc_status.dwCheckPoint = checkpoint++;

	return SetServiceStatus(svc_handle, &svc_status);
}


static VOID WINAPI nt_service_ctrl(DWORD dwCtrlCode)
{
	switch (dwCtrlCode) {
		case SERVICE_CONTROL_STOP:
			ReportStatusToSCMgr(SERVICE_STOP_PENDING, NO_ERROR, 0);
			server_terminate("service is stopped");
			return;
	}
}


static void WINAPI nt_service_main(int argc, char **argv)
{
	/* unexpected to happen, but still.. */
	if (argc < 1)
		return;

	service_name = *argv;

	registry_restore(service_name);
	hEventSource = RegisterEventSource(NULL, "Faerion IRC Daemon");

	if (svc_handle = RegisterServiceCtrlHandler(service_name,
	nt_service_ctrl)) {
		svc_status.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
		svc_status.dwServiceSpecificExitCode = svc_error;

		/* maybe change it to SERVICE_START_PENDING and report
		 * as running from within ircd_main()? */
		if (ReportStatusToSCMgr(SERVICE_RUNNING,
		NO_ERROR, 3000))
			child_monitor();
	}

	/* clean up */
	if (svc_handle)
		ReportStatusToSCMgr(SERVICE_STOPPED, svc_error, 0);

	if (hEventSource)
		DeregisterEventSource(hEventSource);
}


void nt_svc_maintain(char *_svc_name, char *svc_sig)
{
	int retcode;
	char *svc_name, *path, *display_name, tmpbuf[_MAX_PATH + 32];
	SC_HANDLE scmh, sch;
	SERVICE_DESCRIPTION svcdesc;

	if (!svc_sig)
		return;

	svc_name = _svc_name ? _svc_name : "faerion";
	display_name = stricmp(svc_name, "faerion")
		? svc_name : "Faerion IRC Daemon";

	if (!strcmp(svc_sig, "runservice")) {
		SERVICE_TABLE_ENTRY dt[] = {
			{ "faerion", (LPSERVICE_MAIN_FUNCTION)nt_service_main },
			{ NULL, NULL }
		};

		svc_error = 0;

		if (!StartServiceCtrlDispatcher(dt)) {
			/* pointless.  add something to EventLog instead. */
			fprintf(stderr, "Failed to start service.\n");
			exit(1);
		}

		exit(0);
	}

	sch = NULL;
	scmh = NULL;
	retcode = 0;

	path = tmpbuf + 1;

	if (!GetModuleFileName(NULL, path, _MAX_PATH)) {
		fprintf(stderr, "Failed to get my own name?!\n");
		exit(1);
	}

	if (strchr(path, 0x20)) {
		strcat(path, "\"");
		*--path = '"';
	}
	strcat(path, " -k runservice");

	if (!(scmh = OpenSCManager(NULL, NULL, SC_MANAGER_CREATE_SERVICE))) {
		fprintf(stderr, "Failed to open Service Control Manager.\n");
		exit(1);
	}

	if (!strcmp(svc_sig, "install")) {
		if (!(sch = CreateService(scmh, svc_name, display_name,
		SC_MANAGER_ALL_ACCESS, SERVICE_WIN32_OWN_PROCESS,
		SERVICE_AUTO_START, SERVICE_ERROR_IGNORE, path, NULL,
		NULL, "Tcpip\0Afd\0", NULL, NULL))) {
			retcode = 1;
			print_last_error();
			goto finish;
		}

		svcdesc.lpDescription = "Faerion IRC Daemon";
		ChangeServiceConfig2(sch, SERVICE_CONFIG_DESCRIPTION, &svcdesc);

		registry_store(svc_name);

		printf("Service \"%s\" installed.\n", svc_name);
	}
	else if (!strcmp(svc_sig, "uninstall")) {
		if (!(sch = OpenService(scmh, svc_name, SERVICE_CHANGE_CONFIG))) {
			retcode = 1;
			goto finish;
		}

		printf("Service \"%s\" uninstalled.\n", svc_name);
	}
	else if (!strcmp(svc_sig, "configure")) {
		if ((sch = OpenService(scmh, svc_name, SERVICE_CHANGE_CONFIG))
		== NULL) {
			retcode = 1;
			goto finish;
		}

		if (!ChangeServiceConfig(sch, SERVICE_WIN32_OWN_PROCESS,
		SERVICE_AUTO_START, SERVICE_ERROR_NORMAL, path, NULL, NULL,
		"Tcpip\0Afd\0", NULL, NULL, display_name)) {
			retcode = 1;
			goto finish;
		}

		registry_store(svc_name);
	}
	else {
		fprintf(stderr, "Unknown signal for -k: %s\n", svc_sig);
		exit(1);
	}

finish:
	if (retcode != 0)
		print_last_error();

	if (sch)
		CloseServiceHandle(sch);

	if (scmh)
		CloseServiceHandle(scmh);

	exit(retcode);
}
